@@ -25,14 +25,15 @@ import sqlalchemy as sa
 from sqlalchemy import orm
 from sqlalchemy.orm import exc
 
+from neutron._i18n import _, _LE, _LI, _LW
 from neutron.common import constants
 from neutron.common import utils
 from neutron import context as ncontext
 from neutron.db import agents_db
+from neutron.db.availability_zone import network as network_az
 from neutron.db import model_base
 from neutron.extensions import agent as ext_agent
 from neutron.extensions import dhcpagentscheduler
-from neutron.i18n import _LE, _LI, _LW
 
 
 LOG = logging.getLogger(__name__)
@@ -118,16 +119,19 @@ class AgentSchedulerDbMixin(agents_db.AgentDbMixin):
                                          original_agent['host'])
         return result
 
-    def setup_agent_status_check(self, function):
-        self.periodic_agent_loop = loopingcall.FixedIntervalLoopingCall(
-            function)
+    def add_agent_status_check(self, function):
+        loop = loopingcall.FixedIntervalLoopingCall(function)
         # TODO(enikanorov): make interval configurable rather than computed
         interval = max(cfg.CONF.agent_down_time // 2, 1)
         # add random initial delay to allow agents to check in after the
         # neutron server first starts. random to offset multiple servers
         initial_delay = random.randint(interval, interval * 2)
-        self.periodic_agent_loop.start(interval=interval,
-            initial_delay=initial_delay)
+        loop.start(interval=interval, initial_delay=initial_delay)
+
+        if hasattr(self, 'periodic_agent_loops'):
+            self.periodic_agent_loops.append(loop)
+        else:
+            self.periodic_agent_loops = [loop]
 
     def agent_dead_limit_seconds(self):
         return cfg.CONF.agent_down_time * 2
@@ -138,11 +142,12 @@ class AgentSchedulerDbMixin(agents_db.AgentDbMixin):
         # detected, sleep for a while to let the agents check in.
         tdelta = timeutils.utcnow() - getattr(self, '_clock_jump_canary',
                                               timeutils.utcnow())
-        if timeutils.total_seconds(tdelta) > cfg.CONF.agent_down_time:
-            LOG.warn(_LW("Time since last %s agent reschedule check has "
-                         "exceeded the interval between checks. Waiting "
-                         "before check to allow agents to send a heartbeat "
-                         "in case there was a clock adjustment."), agent_type)
+        if tdelta.total_seconds() > cfg.CONF.agent_down_time:
+            LOG.warning(_LW("Time since last %s agent reschedule check has "
+                            "exceeded the interval between checks. Waiting "
+                            "before check to allow agents to send a heartbeat "
+                            "in case there was a clock adjustment."),
+                        agent_type)
             time.sleep(agent_dead_limit)
         self._clock_jump_canary = timeutils.utcnow()
 
@@ -166,7 +171,7 @@ class DhcpAgentSchedulerDbMixin(dhcpagentscheduler
                          "automatic network rescheduling is disabled."))
             return
 
-        self.setup_agent_status_check(self.remove_networks_from_down_agents)
+        self.add_agent_status_check(self.remove_networks_from_down_agents)
 
     def is_eligible_agent(self, context, active, agent):
         # eligible agent is active or starting up
@@ -263,14 +268,13 @@ class DhcpAgentSchedulerDbMixin(dhcpagentscheduler
         cutoff = self.get_cutoff_time(agent_dead_limit)
 
         context = ncontext.get_admin_context()
-        down_bindings = (
-            context.session.query(NetworkDhcpAgentBinding).
-            join(agents_db.Agent).
-            filter(agents_db.Agent.heartbeat_timestamp < cutoff,
-                   agents_db.Agent.admin_state_up))
-        dhcp_notifier = self.agent_notifiers.get(constants.AGENT_TYPE_DHCP)
-
         try:
+            down_bindings = (
+                context.session.query(NetworkDhcpAgentBinding).
+                join(agents_db.Agent).
+                filter(agents_db.Agent.heartbeat_timestamp < cutoff,
+                       agents_db.Agent.admin_state_up))
+            dhcp_notifier = self.agent_notifiers.get(constants.AGENT_TYPE_DHCP)
             dead_bindings = [b for b in
                              self._filter_bindings(context, down_bindings)]
             agents = self.get_agents_db(
@@ -278,17 +282,17 @@ class DhcpAgentSchedulerDbMixin(dhcpagentscheduler
             active_agents = [agent for agent in agents if
                              self.is_eligible_agent(context, True, agent)]
             if not active_agents:
-                LOG.warn(_LW("No DHCP agents available, "
-                             "skipping rescheduling"))
+                LOG.warning(_LW("No DHCP agents available, "
+                                "skipping rescheduling"))
                 return
             for binding in dead_bindings:
-                LOG.warn(_LW("Removing network %(network)s from agent "
-                             "%(agent)s because the agent did not report "
-                             "to the server in the last %(dead_time)s "
-                             "seconds."),
-                         {'network': binding.network_id,
-                          'agent': binding.dhcp_agent_id,
-                          'dead_time': agent_dead_limit})
+                LOG.warning(_LW("Removing network %(network)s from agent "
+                                "%(agent)s because the agent did not report "
+                                "to the server in the last %(dead_time)s "
+                                "seconds."),
+                            {'network': binding.network_id,
+                             'agent': binding.dhcp_agent_id,
+                             'dead_time': agent_dead_limit})
                 # save binding object to avoid ObjectDeletedError
                 # in case binding is concurrently deleted from the DB
                 saved_binding = {'net': binding.network_id,
@@ -456,6 +460,15 @@ class DhcpAgentSchedulerDbMixin(dhcpagentscheduler
             self.network_scheduler.auto_schedule_networks(self, context, host)
 
 
+class AZDhcpAgentSchedulerDbMixin(DhcpAgentSchedulerDbMixin,
+                                  network_az.NetworkAvailabilityZoneMixin):
+    """Mixin class to add availability_zone supported DHCP agent scheduler."""
+
+    def get_network_availability_zones(self, network):
+        zones = {agent.availability_zone for agent in network.dhcp_agents}
+        return list(zones)
+
+
 # helper functions for readability.
 def services_available(admin_state_up):
     if cfg.CONF.enable_services_on_agents_with_admin_state_down:
