@@ -39,16 +39,28 @@ from neutron.agent.linux import ip_lib
 from neutron.agent.linux import utils
 from neutron.common import constants as n_const
 from neutron.db import db_base_plugin_common
+from neutron.plugins.ml2.drivers.linuxbridge.agent import \
+    linuxbridge_neutron_agent as linuxbridge_agent
 from neutron.tests import base as tests_base
 from neutron.tests.common import base as common_base
 from neutron.tests import tools
 
+UNDEFINED = object()
+
 NS_PREFIX = 'test-'
 BR_PREFIX = 'test-br'
-PORT_PREFIX = 'test-port'
+PORT_PREFIX = 'port'
 VETH0_PREFIX = 'test-veth0'
 VETH1_PREFIX = 'test-veth1'
 PATCH_PREFIX = 'patch'
+MACVTAP_PREFIX = 'macvtap'
+
+# port name should be shorter than DEVICE_NAME_MAX_LEN because if this
+# port is used to provide vlan connection between two linuxbridge
+# agents then place for vlan ID is also required, Vlan ID can take max 4 digits
+# and there is also additional "." in device name so it will in overall gives
+# DEVICE_NAME_MAX_LEN = 15 chars
+LB_DEVICE_NAME_MAX_LEN = 10
 
 SS_SOURCE_PORT_PATTERN = re.compile(
     r'^.*\s+\d+\s+.*:(?P<port>\d+)\s+[0-9:].*')
@@ -250,13 +262,102 @@ class RootHelperProcess(subprocess.Popen):
         self.child_pid = utils.get_root_helper_child_pid(
             self.pid, run_as_root=True)
 
+    @property
+    def is_running(self):
+        return self.poll() is None
+
+
+class Pinger(object):
+    """Class for sending ICMP packets asynchronously
+
+    The aim is to keep sending ICMP packets on background while executing other
+    code. After background 'ping' command is stopped, statistics are available.
+
+    Difference to assert_(no_)ping() functions located in this module is that
+    these methods send given count of ICMP packets while they wait for the
+    exit code of 'ping' command.
+
+    >>> pinger = Pinger('pinger_test', '192.168.0.2')
+
+    >>> pinger.start(); time.sleep(5); pinger.stop()
+
+    >>> pinger.sent, pinger.received
+    7 7
+
+    """
+
+    stats_pattern = re.compile(
+        r'^(?P<trans>\d+) packets transmitted,.*(?P<recv>\d+) received.*$')
+    unreachable_pattern = re.compile(
+        r'.* Destination .* Unreachable')
+    TIMEOUT = 15
+
+    def __init__(self, namespace, address, count=None, timeout=1):
+        self.proc = None
+        self.namespace = namespace
+        self.address = address
+        self.count = count
+        self.timeout = timeout
+        self.destination_unreachable = False
+        self.sent = 0
+        self.received = 0
+
+    def _wait_for_death(self):
+        is_dead = lambda: self.proc.poll() is not None
+        utils.wait_until_true(
+            is_dead, timeout=self.TIMEOUT, exception=RuntimeError(
+                "Ping command hasn't ended after %d seconds." % self.TIMEOUT))
+
+    def _parse_stats(self):
+        for line in self.proc.stdout:
+            if (not self.destination_unreachable and
+                    self.unreachable_pattern.match(line)):
+                self.destination_unreachable = True
+                continue
+            result = self.stats_pattern.match(line)
+            if result:
+                self.sent = int(result.group('trans'))
+                self.received = int(result.group('recv'))
+                break
+        else:
+            raise RuntimeError("Didn't find ping statistics.")
+
+    def start(self):
+        if self.proc and self.proc.is_running:
+            raise RuntimeError("This pinger has already a running process")
+        ip_version = ip_lib.get_ip_version(self.address)
+        ping_exec = 'ping' if ip_version == 4 else 'ping6'
+        cmd = [ping_exec, self.address, '-W', str(self.timeout)]
+        if self.count:
+            cmd.extend(['-c', str(self.count)])
+        self.proc = RootHelperProcess(cmd, namespace=self.namespace)
+
+    def stop(self):
+        if self.proc and self.proc.is_running:
+            self.proc.kill(signal.SIGINT)
+            self._wait_for_death()
+            self._parse_stats()
+
+    def wait(self):
+        if self.count:
+            self._wait_for_death()
+            self._parse_stats()
+        else:
+            raise RuntimeError("Pinger is running infinitelly, use stop() "
+                               "first")
+
 
 class NetcatTester(object):
     TCP = n_const.PROTO_NAME_TCP
     UDP = n_const.PROTO_NAME_UDP
+    VERSION_TO_ALL_ADDRESS = {
+        4: '0.0.0.0',
+        6: '::',
+    }
 
     def __init__(self, client_namespace, server_namespace, address,
-                 dst_port, protocol, server_address='0.0.0.0', src_port=None):
+                 dst_port, protocol, server_address=None, src_port=None):
+
         """
         Tool for testing connectivity on transport layer using netcat
         executable.
@@ -282,12 +383,14 @@ class NetcatTester(object):
         self._client_process = None
         self._server_process = None
         self.address = address
-        self.server_address = server_address
         self.dst_port = str(dst_port)
         self.src_port = str(src_port) if src_port else None
         if protocol not in TRANSPORT_PROTOCOLS:
             raise ValueError("Unsupported protocol %s" % protocol)
         self.protocol = protocol
+        ip_version = netaddr.IPAddress(address).version
+        self.server_address = (
+            server_address or self.VERSION_TO_ALL_ADDRESS[ip_version])
 
     @property
     def client_process(self):
@@ -312,7 +415,7 @@ class NetcatTester(object):
         return bool(self._client_process and not self._client_process.poll())
 
     def establish_connection(self):
-        if self._client_process:
+        if self.is_established:
             raise RuntimeError('%(proto)s connection to %(ip_addr)s is already'
                                ' established' %
                                {'proto': self.protocol,
@@ -325,7 +428,7 @@ class NetcatTester(object):
             address=self.address)
         if self.protocol == self.UDP:
             # Create an ASSURED entry in conntrack table for UDP packets,
-            # that requires 3-way communcation
+            # that requires 3-way communication
             # 1st transmission creates UNREPLIED
             # 2nd transmission removes UNREPLIED
             # 3rd transmission creates ASSURED
@@ -402,7 +505,7 @@ class VethFixture(fixtures.Fixture):
     """Create a veth.
 
     :ivar ports: created veth ports
-    :type ports: IPDevice 2-uplet
+    :type ports: tuple of 2 IPDevice
     """
 
     def _setUp(self):
@@ -417,13 +520,15 @@ class VethFixture(fixtures.Fixture):
     def destroy(self):
         for port in self.ports:
             ip_wrapper = ip_lib.IPWrapper(port.namespace)
-            try:
-                ip_wrapper.del_veth(port.name)
-                break
-            except RuntimeError:
-                # NOTE(cbrandily): It seems a veth is automagically deleted
-                # when a namespace owning a veth endpoint is deleted.
-                pass
+            if (ip_wrapper.netns.exists(port.namespace) or
+                port.namespace is None):
+                try:
+                    ip_wrapper.del_veth(port.name)
+                    break
+                except RuntimeError:
+                    # NOTE(cbrandily): It seems a veth is automagically deleted
+                    # when a namespace owning a veth endpoint is deleted.
+                    pass
 
     @staticmethod
     def get_peer_name(name):
@@ -435,6 +540,66 @@ class VethFixture(fixtures.Fixture):
             tools.fail('%s is not a valid VethFixture veth endpoint' % name)
 
 
+class NamedVethFixture(VethFixture):
+    """Create a veth with at least one specified name of a device
+
+    :ivar ports: created veth ports
+    :type ports: tuple of 2 IPDevice
+    """
+
+    def __init__(self, veth0_prefix=VETH0_PREFIX, veth1_prefix=VETH1_PREFIX):
+        super(NamedVethFixture, self).__init__()
+        self.veth0_name = self.get_veth_name(veth0_prefix)
+        self.veth1_name = self.get_veth_name(veth1_prefix)
+
+    def _setUp(self):
+        ip_wrapper = ip_lib.IPWrapper()
+        self.ports = ip_wrapper.add_veth(self.veth0_name, self.veth1_name)
+        self.addCleanup(self.destroy)
+
+    @staticmethod
+    def get_veth_name(name):
+        if name.startswith(VETH0_PREFIX):
+            return tests_base.get_rand_device_name(VETH0_PREFIX)
+        if name.startswith(VETH1_PREFIX):
+            return tests_base.get_rand_device_name(VETH1_PREFIX)
+        return name
+
+
+class MacvtapFixture(fixtures.Fixture):
+    """Create a macvtap.
+
+    :param src_dev: source device for macvtap
+    :type src_dev: IPDevice
+    :param mode: mode of macvtap
+    :type mode: string
+    :ivar ip_dev: created macvtap
+    :type ip_dev: IPDevice
+    """
+    def __init__(self, src_dev=None, mode=None, prefix=MACVTAP_PREFIX):
+        super(MacvtapFixture, self).__init__()
+        self.src_dev = src_dev
+        self.mode = mode
+        self.prefix = prefix
+
+    def _setUp(self):
+        ip_wrapper = ip_lib.IPWrapper()
+        self.ip_dev = common_base.create_resource(
+            self.prefix,
+            ip_wrapper.add_macvtap,
+            self.src_dev, mode=self.mode)
+        self.addCleanup(self.destroy)
+
+    def destroy(self):
+        ip_wrapper = ip_lib.IPWrapper(self.ip_dev.namespace)
+        if (ip_wrapper.netns.exists(self.ip_dev.namespace) or
+            self.ip_dev.namespace is None):
+            try:
+                self.ip_dev.link.delete()
+            except RuntimeError:
+                pass
+
+
 @six.add_metaclass(abc.ABCMeta)
 class PortFixture(fixtures.Fixture):
     """Create a port.
@@ -468,7 +633,7 @@ class PortFixture(fixtures.Fixture):
         if isinstance(bridge, ovs_lib.OVSBridge):
             return OVSPortFixture(bridge, namespace, mac, port_id)
         if isinstance(bridge, bridge_lib.BridgeDevice):
-            return LinuxBridgePortFixture(bridge, namespace)
+            return LinuxBridgePortFixture(bridge, namespace, mac, port_id)
         if isinstance(bridge, VethBridge):
             return VethPortFixture(bridge, namespace)
         tools.fail('Unexpected bridge type: %s' % type(bridge))
@@ -505,7 +670,13 @@ class OVSPortFixture(PortFixture):
         interface_config.register_opts(interface.OPTS)
         ovs_interface = interface.OVSInterfaceDriver(interface_config)
 
-        port_name = tests_base.get_rand_device_name(PORT_PREFIX)
+        # because in some tests this port can be used to providing connection
+        # between linuxbridge agents and vlan_id can be also added to this
+        # device name it has to be max LB_DEVICE_NAME_MAX_LEN long
+        port_name = tests_base.get_rand_name(
+            LB_DEVICE_NAME_MAX_LEN,
+            PORT_PREFIX
+        )
         ovs_interface.plug_new(
             None,
             self.port_id,
@@ -525,16 +696,44 @@ class LinuxBridgeFixture(fixtures.Fixture):
     :ivar namespace: created bridge namespace
     :type namespace: str
     """
+    def __init__(self, prefix=BR_PREFIX, namespace=UNDEFINED,
+                 prefix_is_full_name=False):
+        super(LinuxBridgeFixture, self).__init__()
+        self.prefix = prefix
+        self.prefix_is_full_name = prefix_is_full_name
+        self.namespace = namespace
 
     def _setUp(self):
-        self.namespace = self.useFixture(NamespaceFixture()).name
-        self.bridge = common_base.create_resource(
-            BR_PREFIX,
-            bridge_lib.BridgeDevice.addbr,
-            namespace=self.namespace)
-        self.addCleanup(self.bridge.delbr)
+        if self.namespace is UNDEFINED:
+            self.namespace = self.useFixture(NamespaceFixture()).name
+        self.bridge = self._create_bridge()
+        self.addCleanup(self.safe_delete)
         self.bridge.link.set_up()
-        self.addCleanup(self.bridge.link.set_down)
+        self.addCleanup(self.safe_set_down)
+
+    def safe_set_down(self):
+        try:
+            self.bridge.link.set_down()
+        except RuntimeError:
+            pass
+
+    def safe_delete(self):
+        try:
+            self.bridge.delbr()
+        except RuntimeError:
+            pass
+
+    def _create_bridge(self):
+        if self.prefix_is_full_name:
+            return bridge_lib.BridgeDevice.addbr(
+                name=self.prefix,
+                namespace=self.namespace
+            )
+        else:
+            return common_base.create_resource(
+                self.prefix,
+                bridge_lib.BridgeDevice.addbr,
+                namespace=self.namespace)
 
 
 class LinuxBridgePortFixture(PortFixture):
@@ -546,12 +745,29 @@ class LinuxBridgePortFixture(PortFixture):
     :type br_port: IPDevice
     """
 
+    def __init__(self, bridge, namespace=None, mac=None, port_id=None):
+        super(LinuxBridgePortFixture, self).__init__(
+            bridge, namespace, mac, port_id)
+        # we need to override port_id value here because in Port() class it is
+        # always generated as random. In LinuxBridgePortFixture we need to have
+        # it empty if it was not give because then proper veth_pair will be
+        # created (for example in some functional tests)
+        self.port_id = port_id
+
     def _create_bridge_fixture(self):
         return LinuxBridgeFixture()
 
     def _setUp(self):
         super(LinuxBridgePortFixture, self)._setUp()
-        self.port, self.br_port = self.useFixture(VethFixture()).ports
+        br_port_name = self._get_port_name()
+        if br_port_name:
+            self.br_port, self.port = self.useFixture(
+                NamedVethFixture(veth0_prefix=br_port_name)).ports
+        else:
+            self.br_port, self.port = self.useFixture(VethFixture()).ports
+
+        if self.mac:
+            self.port.link.set_address(self.mac)
 
         # bridge side
         br_ip_wrapper = ip_lib.IPWrapper(self.bridge.namespace)
@@ -564,6 +780,12 @@ class LinuxBridgePortFixture(PortFixture):
         ns_ip_wrapper.add_device_to_namespace(self.port)
         self.port.link.set_up()
 
+    def _get_port_name(self):
+        if self.port_id:
+            return linuxbridge_agent.LinuxBridgeManager.get_tap_device_name(
+                self.port_id)
+        return None
+
 
 class VethBridge(object):
 
