@@ -14,27 +14,32 @@
 #    under the License.
 
 import functools
+import uuid
 
 import fixtures
 import mock
 import six
 import testtools
-import uuid
 import webob
 
 from oslo_db import exception as db_exc
 from oslo_utils import uuidutils
 from sqlalchemy.orm import exc as sqla_exc
 
+from neutron._i18n import _
+from neutron.callbacks import events
 from neutron.callbacks import registry
+from neutron.callbacks import resources
 from neutron.common import constants
 from neutron.common import exceptions as exc
 from neutron.common import utils
 from neutron import context
+from neutron.db import agents_db
 from neutron.db import api as db_api
 from neutron.db import db_base_plugin_v2 as base_plugin
 from neutron.db import l3_db
 from neutron.db import models_v2
+from neutron.extensions import availability_zone as az_ext
 from neutron.extensions import external_net
 from neutron.extensions import multiprovidernet as mpnet
 from neutron.extensions import portbindings
@@ -55,6 +60,7 @@ from neutron.tests.unit import _test_extension_portbindings as test_bindings
 from neutron.tests.unit.agent import test_securitygroups_rpc as test_sg_rpc
 from neutron.tests.unit.db import test_allowedaddresspairs_db as test_pair
 from neutron.tests.unit.db import test_db_base_plugin_v2 as test_plugin
+from neutron.tests.unit.db import test_ipam_pluggable_backend as test_ipam
 from neutron.tests.unit.extensions import test_extra_dhcp_opt as test_dhcpopts
 from neutron.tests.unit.plugins.ml2.drivers import mechanism_logger as \
      mech_logger
@@ -68,7 +74,7 @@ config.cfg.CONF.import_opt('network_vlan_ranges',
 
 PLUGIN_NAME = 'neutron.plugins.ml2.plugin.Ml2Plugin'
 
-DEVICE_OWNER_COMPUTE = 'compute:None'
+DEVICE_OWNER_COMPUTE = constants.DEVICE_OWNER_COMPUTE_PREFIX + 'fake'
 HOST = 'fake_host'
 
 
@@ -312,15 +318,15 @@ class TestMl2NetworksV2(test_plugin.TestNetworksV2,
 
     def test_create_network_segment_allocation_fails(self):
         plugin = manager.NeutronManager.get_plugin()
-        with mock.patch.object(plugin.type_manager, 'create_network_segments',
-            side_effect=db_exc.RetryRequest(ValueError())) as f:
-            self.assertRaises(ValueError,
-                              plugin.create_network,
-                              context.get_admin_context(),
-                              {'network': {'tenant_id': 'sometenant',
-                                           'name': 'dummy',
-                                           'admin_state_up': True,
-                                           'shared': False}})
+        with mock.patch.object(
+            plugin.type_manager, 'create_network_segments',
+            side_effect=db_exc.RetryRequest(ValueError())
+        ) as f:
+            data = {'network': {'tenant_id': 'sometenant', 'name': 'dummy',
+                                'admin_state_up': True, 'shared': False}}
+            req = self.new_create_request('networks', data)
+            res = req.get_response(self.api)
+            self.assertEqual(500, res.status_int)
             self.assertEqual(db_api.MAX_RETRIES + 1, f.call_count)
 
 
@@ -372,7 +378,7 @@ class TestExternalNetwork(Ml2PluginV2TestCase):
 class TestMl2NetworksWithVlanTransparencyAndMTU(TestMl2NetworksV2):
     def setUp(self, plugin=None):
         config.cfg.CONF.set_override('path_mtu', 1000, group='ml2')
-        config.cfg.CONF.set_override('segment_mtu', 1000, group='ml2')
+        config.cfg.CONF.set_override('global_physnet_mtu', 1000)
         config.cfg.CONF.set_override('advertise_mtu', True)
         config.cfg.CONF.set_override('vlan_transparent', True)
         super(TestMl2NetworksWithVlanTransparencyAndMTU, self).setUp(plugin)
@@ -387,10 +393,25 @@ class TestMl2NetworksWithVlanTransparencyAndMTU(TestMl2NetworksV2):
         res = network_req.get_response(self.api)
         self.assertEqual(201, res.status_int)
         network = self.deserialize(self.fmt, res)['network']
-        self.assertEqual(network['mtu'], 1000)
+        self.assertEqual(1000, network['mtu'])
         self.assertIn('vlan_transparent', network)
 
 
+class TestMl2NetworksWithAvailabilityZone(TestMl2NetworksV2):
+    def test_create_network_availability_zone(self):
+        az_hints = ['az1', 'az2']
+        data = {'network': {'name': 'net1',
+                            az_ext.AZ_HINTS: az_hints,
+                            'tenant_id': 'tenant_one'}}
+        with mock.patch.object(agents_db.AgentAvailabilityZoneMixin,
+                               'validate_availability_zones'):
+            network_req = self.new_create_request('networks', data)
+            res = network_req.get_response(self.api)
+            self.assertEqual(201, res.status_int)
+            network = self.deserialize(self.fmt, res)['network']
+            self.assertEqual(az_hints, network[az_ext.AZ_HINTS])
+
+
 class TestMl2SubnetsV2(test_plugin.TestSubnetsV2,
                        Ml2PluginV2TestCase):
     def test_delete_subnet_race_with_dhcp_port_creation(self):
@@ -434,6 +455,49 @@ class TestMl2SubnetsV2(test_plugin.TestSubnetsV2,
                     self.assertEqual(204, res.status_int)
 
 
+class TestMl2DbOperationBounds(test_plugin.DbOperationBoundMixin,
+                               Ml2PluginV2TestCase):
+    """Test cases to assert constant query count for list operations.
+
+    These test cases assert that an increase in the number of objects
+    does not result in an increase of the number of db operations. All
+    database lookups during a list operation should be performed in bulk
+    so the number of queries required for 2 objects instead of 1 should
+    stay the same.
+    """
+
+    def setUp(self):
+        super(TestMl2DbOperationBounds, self).setUp()
+        self.kwargs = self.get_api_kwargs()
+
+    def make_network(self):
+        return self._make_network(self.fmt, 'name', True, **self.kwargs)
+
+    def make_subnet(self):
+        net = self.make_network()
+        setattr(self, '_subnet_count', getattr(self, '_subnet_count', 0) + 1)
+        cidr = '1.%s.0.0/24' % self._subnet_count
+        return self._make_subnet(self.fmt, net, None, cidr, **self.kwargs)
+
+    def make_port(self):
+        net = self.make_network()
+        return self._make_port(self.fmt, net['network']['id'], **self.kwargs)
+
+    def test_network_list_queries_constant(self):
+        self._assert_object_list_queries_constant(self.make_network,
+                                                  'networks')
+
+    def test_subnet_list_queries_constant(self):
+        self._assert_object_list_queries_constant(self.make_subnet, 'subnets')
+
+    def test_port_list_queries_constant(self):
+        self._assert_object_list_queries_constant(self.make_port, 'ports')
+
+
+class TestMl2DbOperationBoundsTenant(TestMl2DbOperationBounds):
+    admin = False
+
+
 class TestMl2PortsV2(test_plugin.TestPortsV2, Ml2PluginV2TestCase):
 
     def test_update_port_status_build(self):
@@ -504,7 +568,8 @@ class TestMl2PortsV2(test_plugin.TestPortsV2, Ml2PluginV2TestCase):
                 l3plugin.create_floatingip(
                     context.get_admin_context(),
                     {'floatingip': {'floating_network_id': n['network']['id'],
-                                    'tenant_id': n['network']['tenant_id']}}
+                                    'tenant_id': n['network']['tenant_id'],
+                                    'dns_name': '', 'dns_domain': ''}}
                 )
         self._delete('networks', n['network']['id'])
         flips = l3plugin.get_floatingips(context.get_admin_context())
@@ -642,7 +707,7 @@ class TestMl2PortsV2(test_plugin.TestPortsV2, Ml2PluginV2TestCase):
             self.assertTrue(notify.call_counts)
 
     def test_check_if_compute_port_serviced_by_dvr(self):
-        self.assertTrue(utils.is_dvr_serviced('compute:None'))
+        self.assertTrue(utils.is_dvr_serviced(DEVICE_OWNER_COMPUTE))
 
     def test_check_if_lbaas_vip_port_serviced_by_dvr(self):
         self.assertTrue(utils.is_dvr_serviced(
@@ -671,18 +736,32 @@ class TestMl2PortsV2(test_plugin.TestPortsV2, Ml2PluginV2TestCase):
             self.assertIsNone(l3plugin.disassociate_floatingips(ctx, port_id))
 
     def test_create_port_tolerates_db_deadlock(self):
-        ctx = context.get_admin_context()
         with self.network() as net:
             with self.subnet(network=net) as subnet:
-                segments = ml2_db.get_network_segments(ctx.session,
-                                                       net['network']['id'])
+                _orig = ml2_db.get_locked_port_and_binding
+                self._failed = False
+
+                def fail_once(*args, **kwargs):
+                    if not self._failed:
+                        self._failed = True
+                        raise db_exc.DBDeadlock()
+                    return _orig(*args, **kwargs)
                 with mock.patch('neutron.plugins.ml2.plugin.'
-                                'db.get_network_segments') as get_seg_mock:
-                    get_seg_mock.side_effect = [db_exc.DBDeadlock, segments,
-                                                segments, segments]
-                    with self.port(subnet=subnet) as port:
+                                'db.get_locked_port_and_binding',
+                                side_effect=fail_once) as get_port_mock:
+                    port_kwargs = {portbindings.HOST_ID: 'host1',
+                                   'subnet': subnet,
+                                   'device_id': 'deadlocktest'}
+                    with self.port(arg_list=(portbindings.HOST_ID,),
+                                   **port_kwargs) as port:
                         self.assertTrue(port['port']['id'])
-                        self.assertEqual(4, get_seg_mock.call_count)
+                        self.assertTrue(get_port_mock.called)
+                        # make sure that we didn't create more than one port on
+                        # the retry
+                        query_params = "network_id=%s" % net['network']['id']
+                        query_params += "&device_id=%s" % 'deadlocktest'
+                        ports = self._list('ports', query_params=query_params)
+                        self.assertEqual(1, len(ports['ports']))
 
     def test_delete_port_tolerates_db_deadlock(self):
         ctx = context.get_admin_context()
@@ -701,10 +780,81 @@ class TestMl2PortsV2(test_plugin.TestPortsV2, Ml2PluginV2TestCase):
                 self.assertRaises(
                     exc.PortNotFound, plugin.get_port, ctx, port['port']['id'])
 
+    def test_port_create_resillient_to_duplicate_records(self):
+
+        def make_port():
+            with self.port():
+                pass
+
+        self._test_operation_resillient_to_ipallocation_failure(make_port)
+
+    def test_port_update_resillient_to_duplicate_records(self):
+        with self.port() as p:
+            data = {'port': {'fixed_ips': [{'ip_address': '10.0.0.9'}]}}
+            req = self.new_update_request('ports', data, p['port']['id'])
+
+            def do_request():
+                self.assertEqual(200, req.get_response(self.api).status_int)
+
+            self._test_operation_resillient_to_ipallocation_failure(do_request)
+
+    def _test_operation_resillient_to_ipallocation_failure(self, func):
+        from sqlalchemy import event
+
+        class IPAllocationsGrenade(object):
+            insert_ip_called = False
+            except_raised = False
+
+            def execute(self, con, curs, stmt, *args, **kwargs):
+                if 'INSERT INTO ipallocations' in stmt:
+                    self.insert_ip_called = True
+
+            def commit(self, con):
+                # we blow up on commit to simulate another thread/server
+                # stealing our IP before our transaction was done
+                if self.insert_ip_called and not self.except_raised:
+                    self.except_raised = True
+                    raise db_exc.DBDuplicateEntry()
+
+        listener = IPAllocationsGrenade()
+        engine = db_api.get_engine()
+        event.listen(engine, 'before_cursor_execute', listener.execute)
+        event.listen(engine, 'commit', listener.commit)
+        self.addCleanup(event.remove, engine, 'before_cursor_execute',
+                        listener.execute)
+        self.addCleanup(event.remove, engine, 'commit',
+                        listener.commit)
+        func()
+        # make sure that the grenade went off during the commit
+        self.assertTrue(listener.except_raised)
+
 
 class TestMl2PluginOnly(Ml2PluginV2TestCase):
     """For testing methods that don't call drivers"""
 
+    def test__verify_service_plugins_requirements(self):
+        plugin = manager.NeutronManager.get_plugin()
+        with mock.patch.dict(ml2_plugin.SERVICE_PLUGINS_REQUIRED_DRIVERS,
+                             {self.l3_plugin: self._mechanism_drivers}),\
+                mock.patch.object(plugin.extension_manager,
+                                  'names',
+                                  return_value=self._mechanism_drivers):
+
+            plugin._verify_service_plugins_requirements()
+
+    def test__verify_service_plugins_requirements_missing_driver(self):
+        plugin = manager.NeutronManager.get_plugin()
+        with mock.patch.dict(ml2_plugin.SERVICE_PLUGINS_REQUIRED_DRIVERS,
+                             {self.l3_plugin: ['test_required_driver']}),\
+                mock.patch.object(plugin.extension_manager,
+                                  'names',
+                                  return_value=self._mechanism_drivers):
+
+            self.assertRaises(
+                ml2_exc.ExtensionDriverNotFound,
+                plugin._verify_service_plugins_requirements
+            )
+
     def _test_check_mac_update_allowed(self, vif_type, expect_change=True):
         plugin = manager.NeutronManager.get_plugin()
         port = {'mac_address': "fake_mac", 'id': "fake_id"}
@@ -766,54 +916,45 @@ class TestMl2DvrPortsV2(TestMl2PortsV2):
             mock.PropertyMock(return_value=extensions))
         self.service_plugins = {'L3_ROUTER_NAT': self.l3plugin}
 
-    def _test_delete_dvr_serviced_port(self, device_owner, floating_ip=False):
+    def test_delete_port_notifies_l3_plugin(self, floating_ip=False):
         ns_to_delete = {'host': 'myhost', 'agent_id': 'vm_l3_agent',
                         'router_id': 'my_router'}
-        fip_set = set()
+        router_ids = set()
         if floating_ip:
-            fip_set.add(ns_to_delete['router_id'])
+            router_ids.add(ns_to_delete['router_id'])
 
         with mock.patch.object(manager.NeutronManager,
                                'get_service_plugins',
                                return_value=self.service_plugins),\
-                self.port(device_owner=device_owner) as port,\
+                self.port() as port,\
                 mock.patch.object(registry, 'notify') as notify,\
                 mock.patch.object(self.l3plugin,
                                   'disassociate_floatingips',
-                                  return_value=fip_set),\
-                mock.patch.object(
-                    self.l3plugin,
-                    'dvr_deletens_if_no_port',
-                    return_value=[ns_to_delete]) as dvr_delns_ifno_port:
-
+                                  return_value=router_ids):
             port_id = port['port']['id']
             self.plugin.delete_port(self.context, port_id)
+            self.assertEqual(2, notify.call_count)
+            # needed for a full match in the assertion below
+            port['port']['extra_dhcp_opts'] = []
+            expected = [mock.call(resources.PORT, events.BEFORE_DELETE,
+                                  mock.ANY, context=self.context,
+                                  port_id=port['port']['id'], port_check=True),
+                        mock.call(resources.PORT, events.AFTER_DELETE,
+                                  mock.ANY, context=self.context,
+                                  port=port['port'],
+                                  router_ids=router_ids)]
+            notify.assert_has_calls(expected)
 
-            self.assertTrue(notify.call_count)
-            dvr_delns_ifno_port.assert_called_once_with(self.context,
-                                                        port['port']['id'])
-
-    def test_delete_last_vm_port(self):
-        self._test_delete_dvr_serviced_port(device_owner='compute:None')
-
-    def test_delete_last_vm_port_with_floatingip(self):
-        self._test_delete_dvr_serviced_port(device_owner='compute:None',
-                                            floating_ip=True)
-
-    def test_delete_lbaas_vip_port(self):
-        self._test_delete_dvr_serviced_port(
-            device_owner=constants.DEVICE_OWNER_LOADBALANCER)
-
-    def test_delete_lbaasv2_vip_port(self):
-        self._test_delete_dvr_serviced_port(
-            device_owner=constants.DEVICE_OWNER_LOADBALANCERV2)
+    def test_delete_port_with_floatingip_notifies_l3_plugin(self):
+        self.test_delete_port_notifies_l3_plugin(floating_ip=True)
 
     def test_concurrent_csnat_port_delete(self):
         plugin = manager.NeutronManager.get_service_plugins()[
             p_const.L3_ROUTER_NAT]
         r = plugin.create_router(
             self.context,
-            {'router': {'name': 'router', 'admin_state_up': True}})
+            {'router': {'name': 'router', 'admin_state_up': True,
+             'tenant_id': self.context.tenant_id}})
         with self.subnet() as s:
             p = plugin.add_router_interface(self.context, r['id'],
                                             {'subnet_id': s['subnet']['id']})
@@ -916,6 +1057,92 @@ class TestMl2PortBinding(Ml2PluginV2TestCase,
             # should have returned before calling _make_port_dict
             self.assertFalse(mpd_mock.mock_calls)
 
+    def _create_port_and_bound_context(self, port_vif_type, bound_vif_type):
+        with self.port() as port:
+            plugin = manager.NeutronManager.get_plugin()
+            binding = ml2_db.get_locked_port_and_binding(self.context.session,
+                                                         port['port']['id'])[1]
+            binding['host'] = 'fake_host'
+            binding['vif_type'] = port_vif_type
+            # Generates port context to be used before the bind.
+            port_context = driver_context.PortContext(
+                plugin, self.context, port['port'],
+                plugin.get_network(self.context, port['port']['network_id']),
+                binding, None)
+            bound_context = mock.MagicMock()
+            # Bound context is how port_context is expected to look
+            # after _bind_port.
+            bound_context.vif_type = bound_vif_type
+            return plugin, port_context, bound_context
+
+    def test__attempt_binding(self):
+        # Simulate a successful binding for vif_type unbound
+        # and keep the same binding state for other vif types.
+        vif_types = [(portbindings.VIF_TYPE_BINDING_FAILED,
+                      portbindings.VIF_TYPE_BINDING_FAILED),
+                     (portbindings.VIF_TYPE_UNBOUND,
+                      portbindings.VIF_TYPE_OVS),
+                     (portbindings.VIF_TYPE_OVS,
+                      portbindings.VIF_TYPE_OVS)]
+
+        for port_vif_type, bound_vif_type in vif_types:
+            plugin, port_context, bound_context = (
+                self._create_port_and_bound_context(port_vif_type,
+                                                    bound_vif_type))
+            with mock.patch('neutron.plugins.ml2.plugin.Ml2Plugin._bind_port',
+                            return_value=bound_context) as bd_mock:
+                context, need_notify, try_again = (plugin._attempt_binding(
+                    port_context, False))
+                expected_need_notify = port_vif_type not in (
+                    portbindings.VIF_TYPE_BINDING_FAILED,
+                    portbindings.VIF_TYPE_OVS)
+
+                if bound_vif_type == portbindings.VIF_TYPE_BINDING_FAILED:
+                    expected_vif_type = port_vif_type
+                    expected_try_again = True
+                    expected_bd_mock_called = True
+                else:
+                    expected_vif_type = portbindings.VIF_TYPE_OVS
+                    expected_try_again = False
+                    expected_bd_mock_called = (port_vif_type ==
+                                               portbindings.VIF_TYPE_UNBOUND)
+
+                self.assertEqual(expected_need_notify, need_notify)
+                self.assertEqual(expected_vif_type, context.vif_type)
+                self.assertEqual(expected_try_again, try_again)
+                self.assertEqual(expected_bd_mock_called, bd_mock.called)
+
+    def test__attempt_binding_retries(self):
+        # Simulate cases of both successful and failed binding states for
+        # vif_type unbound
+        vif_types = [(portbindings.VIF_TYPE_UNBOUND,
+                      portbindings.VIF_TYPE_BINDING_FAILED),
+                     (portbindings.VIF_TYPE_UNBOUND,
+                      portbindings.VIF_TYPE_OVS)]
+
+        for port_vif_type, bound_vif_type in vif_types:
+            plugin, port_context, bound_context = (
+                self._create_port_and_bound_context(port_vif_type,
+                                                    bound_vif_type))
+            with mock.patch(
+                    'neutron.plugins.ml2.plugin.Ml2Plugin._bind_port',
+                    return_value=bound_context),\
+                    mock.patch('neutron.plugins.ml2.plugin.Ml2Plugin._commit_'
+                               'port_binding',
+                               return_value=(bound_context, True, False)),\
+                    mock.patch('neutron.plugins.ml2.plugin.Ml2Plugin.'
+                               '_attempt_binding',
+                               side_effect=plugin._attempt_binding) as at_mock:
+                    plugin._bind_port_if_needed(port_context)
+                    if bound_vif_type == portbindings.VIF_TYPE_BINDING_FAILED:
+                        # An unsuccessful binding attempt should be retried
+                        # MAX_BIND_TRIES amount of times.
+                        self.assertEqual(ml2_plugin.MAX_BIND_TRIES,
+                                         at_mock.call_count)
+                    else:
+                        # Successful binding should only be attempted once.
+                        self.assertEqual(1, at_mock.call_count)
+
     def test_port_binding_profile_not_changed(self):
         profile = {'e': 5}
         profile_arg = {portbindings.PROFILE: profile}
@@ -989,11 +1216,23 @@ class TestMl2PortBinding(Ml2PluginV2TestCase,
                                  mech_context._binding.router_id)
                 self.assertEqual(host_id, mech_context._binding.host)
 
+    def test_update_dvr_port_binding_on_concurrent_port_delete(self):
+        plugin = manager.NeutronManager.get_plugin()
+        with self.port() as port:
+            port = {
+                'id': port['port']['id'],
+                portbindings.HOST_ID: 'foo_host',
+            }
+            with mock.patch.object(plugin, 'get_port', new=plugin.delete_port):
+                res = plugin.update_dvr_port_binding(
+                    self.context, 'foo_port_id', {'port': port})
+        self.assertIsNone(res)
+
     def test_update_dvr_port_binding_on_non_existent_port(self):
         plugin = manager.NeutronManager.get_plugin()
         port = {
             'id': 'foo_port_id',
-            'binding:host_id': 'foo_host',
+            portbindings.HOST_ID: 'foo_host',
         }
         with mock.patch.object(ml2_db, 'ensure_dvr_port_binding') as mock_dvr:
             plugin.update_dvr_port_binding(
@@ -1249,8 +1488,8 @@ class TestMultiSegmentNetworks(Ml2PluginV2TestCase):
             req = self.new_delete_request('networks', network_id)
             res = req.get_response(self.api)
             self.assertEqual(2, rs.call_count)
-        self.assertEqual(ml2_db.get_network_segments(
-            self.context.session, network_id), [])
+        self.assertEqual([], ml2_db.get_network_segments(
+            self.context.session, network_id))
         self.assertIsNone(ml2_db.get_dynamic_segment(
             self.context.session, network_id, 'physnet2'))
 
@@ -1600,7 +1839,7 @@ class TestFaultyMechansimDriver(Ml2PluginV2FaultyDriverTestCase):
                             network['network']['tenant_id'],
                             'name': 'port1',
                             'device_owner':
-                            'network:router_interface_distributed',
+                            constants.DEVICE_OWNER_DVR_INTERFACE,
                             'admin_state_up': 1,
                             'fixed_ips':
                             [{'subnet_id': subnet_id}]}}
@@ -1621,6 +1860,29 @@ class TestFaultyMechansimDriver(Ml2PluginV2FaultyDriverTestCase):
                         self.assertEqual(new_name, port['port']['name'])
 
 
+class TestML2PluggableIPAM(test_ipam.UseIpamMixin, TestMl2SubnetsV2):
+    def test_create_subnet_delete_subnet_call_ipam_driver(self):
+        driver = 'neutron.ipam.drivers.neutrondb_ipam.driver.NeutronDbPool'
+        gateway_ip = '10.0.0.1'
+        cidr = '10.0.0.0/24'
+        with mock.patch(driver) as driver_mock:
+            request = mock.Mock()
+            request.subnet_id = uuidutils.generate_uuid()
+            request.subnet_cidr = cidr
+            request.allocation_pools = []
+            request.gateway_ip = gateway_ip
+            request.tenant_id = uuidutils.generate_uuid()
+
+            ipam_subnet = mock.Mock()
+            ipam_subnet.get_details.return_value = request
+            driver_mock().allocate_subnet.return_value = ipam_subnet
+
+            self._test_create_subnet(gateway_ip=gateway_ip, cidr=cidr)
+
+            driver_mock().allocate_subnet.assert_called_with(mock.ANY)
+            driver_mock().remove_subnet.assert_called_with(request.subnet_id)
+
+
 class TestMl2PluginCreateUpdateDeletePort(base.BaseTestCase):
     def setUp(self):
         super(TestMl2PluginCreateUpdateDeletePort, self).setUp()
@@ -1634,91 +1896,37 @@ class TestMl2PluginCreateUpdateDeletePort(base.BaseTestCase):
         exit = transaction.__exit__.call_count
         self.assertEqual(enter, exit)
 
-    def _create_plugin_for_create_update_port(self, new_host_port):
+    def _create_plugin_for_create_update_port(self):
         plugin = ml2_plugin.Ml2Plugin()
         plugin.extension_manager = mock.Mock()
         plugin.type_manager = mock.Mock()
         plugin.mechanism_manager = mock.Mock()
         plugin.notifier = mock.Mock()
-        plugin._get_host_port_if_changed = mock.Mock(
-            return_value=new_host_port)
         plugin._check_mac_update_allowed = mock.Mock(return_value=True)
+        plugin._extend_availability_zone = mock.Mock()
 
         self.notify.side_effect = (
             lambda r, e, t, **kwargs: self._ensure_transaction_is_closed())
 
         return plugin
 
-    def _test__get_host_port_if_changed(
-        self, mech_context, attrs=None, expected_retval=None):
-        with mock.patch.object(ml2_plugin.Ml2Plugin,
-                               '__init__',
-                               return_value=None):
-            plugin = ml2_plugin.Ml2Plugin()
-            test_return = plugin._get_host_port_if_changed(mech_context, attrs)
-            self.assertEqual(expected_retval, test_return)
-
-    def test__get_host_port_if_changed_no_attrs(self):
-        mech_context = mock.Mock()
-        mech_context._binding.host = 'Host-1'
-        self._test__get_host_port_if_changed(
-            mech_context, attrs=None, expected_retval=None)
-
-    def test__get_host_port_if_changed_no_binding_change(self):
-        mech_context = mock.Mock()
-        mech_context._binding.host = 'Host-1'
-        mech_context.current = {
-            'id': 'fake-id',
-            'mac_address': '2a:2b:2c:2d:2e:2f'
-        }
-        attrs = {'mac_address': '0a:0b:0c:0d:0e:0f'}
-        self._test__get_host_port_if_changed(
-            mech_context, attrs=attrs, expected_retval=None)
-
-        attrs = {
-            portbindings.HOST_ID: 'Host-1',
-            'mac_address': '0a:0b:0c:0d:0e:0f',
-        }
-        self._test__get_host_port_if_changed(
-            mech_context, attrs=attrs, expected_retval=None)
-
-    def test__get_host_port_if_changed_with_binding_removed(self):
-        expected_return = {
-            'id': 'fake-id',
-            portbindings.HOST_ID: None,
-            'mac_address': '2a:2b:2c:2d:2e:2f'
-        }
-        mech_context = mock.Mock()
-        mech_context._binding.host = 'Host-1'
-        mech_context.current = expected_return
-        attrs = {portbindings.HOST_ID: None}
-        self._test__get_host_port_if_changed(
-            mech_context, attrs=attrs, expected_retval=expected_return)
-
-    def test__get_host_port_if_changed_with_binding_added(self):
-        expected_return = {
-            'id': 'fake-id',
-            portbindings.HOST_ID: 'host-1',
-            'mac_address': '2a:2b:2c:2d:2e:2f'
-        }
-        mech_context = mock.Mock()
-        mech_context.current = expected_return
-        attrs = {portbindings.HOST_ID: 'host-1'}
-        self._test__get_host_port_if_changed(
-            mech_context, attrs=attrs, expected_retval=expected_return)
-
     def test_create_port_rpc_outside_transaction(self):
         with mock.patch.object(ml2_plugin.Ml2Plugin, '__init__') as init,\
                 mock.patch.object(base_plugin.NeutronDbPluginV2,
-                                  'create_port'):
+                                  '_make_port_dict') as make_port, \
+                mock.patch.object(base_plugin.NeutronDbPluginV2,
+                                  'update_port'),\
+                mock.patch.object(base_plugin.NeutronDbPluginV2,
+                                  'create_port_db'):
             init.return_value = None
 
-            new_host_port = mock.Mock()
-            plugin = self._create_plugin_for_create_update_port(new_host_port)
+            new_port = mock.MagicMock()
+            make_port.return_value = new_port
+            plugin = self._create_plugin_for_create_update_port()
 
             plugin.create_port(self.context, mock.MagicMock())
 
-            kwargs = {'context': self.context, 'port': new_host_port}
+            kwargs = {'context': self.context, 'port': new_port}
             self.notify.assert_called_once_with('port', 'after_create',
                 plugin, **kwargs)
 
@@ -1733,7 +1941,7 @@ class TestMl2PluginCreateUpdateDeletePort(base.BaseTestCase):
             admin_state_up=True,
             status='ACTIVE',
             device_id='vm_id',
-            device_owner='compute:None')
+            device_owner=DEVICE_OWNER_COMPUTE)
 
         binding = mock.Mock()
         binding.port_id = port_id
@@ -1747,17 +1955,18 @@ class TestMl2PluginCreateUpdateDeletePort(base.BaseTestCase):
                 mock.patch.object(ml2_db, 'get_locked_port_and_binding',
                                   return_value=(original_port_db, binding)),\
                 mock.patch.object(base_plugin.NeutronDbPluginV2,
-                                  'update_port'):
+                                  'update_port') as db_update_port:
             init.return_value = None
-            new_host_port = mock.Mock()
-            plugin = self._create_plugin_for_create_update_port(new_host_port)
+            updated_port = mock.MagicMock()
+            db_update_port.return_value = updated_port
+            plugin = self._create_plugin_for_create_update_port()
             original_port = plugin._make_port_dict(original_port_db)
 
             plugin.update_port(self.context, port_id, mock.MagicMock())
 
             kwargs = {
                 'context': self.context,
-                'port': new_host_port,
+                'port': updated_port,
                 'mac_address_updated': True,
                 'original_port': original_port,
             }
@@ -1778,10 +1987,26 @@ class TestMl2PluginCreateUpdateDeletePort(base.BaseTestCase):
                 mock.patch.object(manager.NeutronManager,
                                   'get_service_plugins',
                                   return_value={'L3_ROUTER_NAT': l3plugin}):
-            plugin = self._create_plugin_for_create_update_port(mock.Mock())
+            plugin = self._create_plugin_for_create_update_port()
             # Set backend manually here since __init__ was mocked
             plugin.set_ipam_backend()
             # deleting the port will call registry.notify, which will
             # run the transaction balancing function defined in this test
             plugin.delete_port(self.context, 'fake_id')
             self.assertTrue(self.notify.call_count)
+
+
+class TestTransactionGuard(Ml2PluginV2TestCase):
+    def test_delete_network_guard(self):
+        plugin = ml2_plugin.Ml2Plugin()
+        ctx = context.get_admin_context()
+        with ctx.session.begin(subtransactions=True):
+            with testtools.ExpectedException(RuntimeError):
+                plugin.delete_network(ctx, 'id')
+
+    def test_delete_subnet_guard(self):
+        plugin = ml2_plugin.Ml2Plugin()
+        ctx = context.get_admin_context()
+        with ctx.session.begin(subtransactions=True):
+            with testtools.ExpectedException(RuntimeError):
+                plugin.delete_subnet(ctx, 'id')
